@@ -8,11 +8,11 @@ static const MVMREPROps MVMIter_this_repr;
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
     MVMSTable *st = MVM_gc_allocate_stable(tc, &MVMIter_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMIter);
-    });
+    }
 
     return st->WHAT;
 }
@@ -105,12 +105,30 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
             }
             return;
         case MVM_ITER_MODE_HASH:
+            ; /* Sigh, C99 won't let me put a declaration here. */
+            MVMStrHashTable *hashtable = &(((MVMHash *)target)->body.hashtable);
+#if HASH_DEBUG_ITER
+            struct MVMStrHashTableControl *control = hashtable->table;
+            MVMuint64 ht_id = control ? control->ht_id : 0;
+            if (body->hash_state.curr.owner != ht_id) {
+                MVM_oops(tc, "MVMIter shift called with an iterator from a different hash table: %016" PRIx64 " != %016" PRIx64,
+                         body->hash_state.curr.owner, ht_id);
+            }
+            /* OK, to implement "delete at current iterator position" we need
+             * to cheat somewhat. */
+            if (MVM_str_hash_iterator_target_deleted(tc, hashtable, body->hash_state.curr)) {
+                /* The only action taken on the hash was to delete at the
+                 * current iterator. In which case, the "next" iterator is
+                 * valid (but has already been advanced beyond pos, so we
+                 * can't perform this test on it). So "fix up" its state to pass
+                 * muster with the HASH_DEBUG_ITER sanity tests. */
+                body->hash_state.next.serial = control->serial;
+            }
+#endif
             body->hash_state.curr = body->hash_state.next;
-            if (!body->hash_state.curr)
+            if (MVM_str_hash_at_end(tc, hashtable, body->hash_state.curr))
                 MVM_exception_throw_adhoc(tc, "Iteration past end of iterator");
-            body->hash_state.next = HASH_ITER_NEXT_ITEM(
-                &(body->hash_state.next->hash_handle),
-                &(body->hash_state.bucket_state));
+            body->hash_state.next = MVM_str_hash_next_nocheck(tc, hashtable, body->hash_state.curr);
             value->o = root;
             return;
         default:
@@ -172,7 +190,9 @@ static const MVMREPROps MVMIter_this_repr = {
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
         get_elem_storage_spec,
         MVM_REPR_DEFAULT_POS_AS_ATOMIC,
-        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM,
+        MVM_REPR_DEFAULT_POS_WRITE_BUF,
+        MVM_REPR_DEFAULT_POS_READ_BUF
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     MVM_REPR_DEFAULT_ELEMS,
@@ -201,7 +221,7 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
     if (!IS_CONCRETE(target)) {
         MVM_exception_throw_adhoc(tc, "Cannot iterate over a %s type object", MVM_6model_get_debug_name(tc, target));
     }
-    MVMROOT(tc, target, {
+    MVMROOT(tc, target) {
         if (REPR(target)->ID == MVM_REPR_ID_VMArray) {
             iterator = (MVMIter *)MVM_repr_alloc_init(tc,
                 MVM_hll_current(tc)->array_iterator_type);
@@ -219,120 +239,21 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
             iterator = (MVMIter *)MVM_repr_alloc_init(tc,
                 MVM_hll_current(tc)->hash_iterator_type);
             iterator->body.mode = MVM_ITER_MODE_HASH;
-            iterator->body.hash_state.bucket_state = 0;
-            iterator->body.hash_state.curr         = NULL;
-            iterator->body.hash_state.next         = HASH_ITER_FIRST_ITEM(
-                ((MVMHash *)target)->body.hash_head
-                    ? ((MVMHash *)target)->body.hash_head->hash_handle.tbl
-                    : NULL,
-                &(iterator->body.hash_state.bucket_state));
+            MVMStrHashTable *hashtable = &(((MVMHash *)target)->body.hashtable);
+            iterator->body.hash_state.curr = MVM_str_hash_start(tc, hashtable);
+            iterator->body.hash_state.next = MVM_str_hash_first(tc, hashtable);
             MVM_ASSIGN_REF(tc, &(iterator->common.header), iterator->body.target, target);
         }
         else if (REPR(target)->ID == MVM_REPR_ID_MVMContext) {
             /* Turn the context into a VMHash and then iterate that. */
-            MVMHLLConfig *hll = MVM_hll_current(tc);
-            MVMObject *ctx_hash = MVM_repr_alloc_init(tc, hll->slurpy_hash_type);
-            MVMROOT(tc, ctx_hash, {
-                MVMContext          *ctx    = (MVMContext *)target;
-                MVMFrame            *frame  = ctx->body.context;
-                MVMStaticFrame      *sf     = frame->static_info;
-                MVMLexicalRegistry **lexreg = sf->body.lexical_names_list;
-                MVMuint32 i;
-                MVMROOT2(tc, frame, sf, {
-                    for (i = 0; i < sf->body.num_lexicals; i++) {
-                        MVMuint32 idx  = lexreg[i]->value;
-                        MVMuint16 type = sf->body.lexical_types[idx];
-                        switch (type) {
-                            case MVM_reg_obj: {
-                                MVMObject *obj = frame->env[idx].o;
-                                if (!obj)
-                                    obj = MVM_frame_vivify_lexical(tc, frame, idx);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, obj);
-                                break;
-                            }
-                            case MVM_reg_str: {
-                                MVMObject *bs = MVM_repr_box_str(tc, hll->str_box_type,
-                                    frame->env[idx].s);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bs);
-                                break;
-                            }
-                            case MVM_reg_int8: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].i8);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_uint8: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].u8);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_int16: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].i16);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_uint16: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].u16);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_int32: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].i32);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_uint32: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].u32);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_int64: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].i64);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_uint64: {
-                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                    frame->env[idx].u64);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                                break;
-                            }
-                            case MVM_reg_num32: {
-                                MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
-                                    frame->env[idx].n32);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
-                                break;
-                            }
-                            case MVM_reg_num64: {
-                                MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
-                                    frame->env[idx].n64);
-                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
-                                break;
-                            }
-                            default:
-                                MVM_exception_throw_adhoc(tc,
-                                    "%s lexical type encountered while building context iterator", MVM_reg_get_debug_name(tc, type));
-                        }
-                    }
-                });
-            });
-
-            /* Call ourselves recursively to get the iterator for this
-            * hash. */
+            MVMObject *ctx_hash = MVM_context_lexicals_as_hash(tc, (MVMContext *)target);
             iterator = (MVMIter *)MVM_iter(tc, ctx_hash);
         }
         else {
             MVM_exception_throw_adhoc(tc, "Cannot iterate object with %s representation (%s)",
                 REPR(target)->name, MVM_6model_get_debug_name(tc, target));
         }
-    });
+    }
     return (MVMObject *)iterator;
 }
 
@@ -342,10 +263,10 @@ MVMint64 MVM_iter_istrue(MVMThreadContext *tc, MVMIter *iter) {
         case MVM_ITER_MODE_ARRAY_INT:
         case MVM_ITER_MODE_ARRAY_NUM:
         case MVM_ITER_MODE_ARRAY_STR:
-            return iter->body.array_state.index + 1 < iter->body.array_state.limit ? 1 : 0;
+            return MVM_iter_istrue_array(tc, iter);
             break;
         case MVM_ITER_MODE_HASH:
-            return iter->body.hash_state.next != NULL ? 1 : 0;
+            return MVM_iter_istrue_hash(tc, iter);
             break;
         default:
             MVM_exception_throw_adhoc(tc, "Invalid iteration mode used");
@@ -356,9 +277,24 @@ MVMString * MVM_iterkey_s(MVMThreadContext *tc, MVMIter *iterator) {
     if (REPR(iterator)->ID != MVM_REPR_ID_MVMIter
             || iterator->body.mode != MVM_ITER_MODE_HASH)
         MVM_exception_throw_adhoc(tc, "This is not a hash iterator, it's a %s (%s)", REPR(iterator)->name, MVM_6model_get_debug_name(tc, (MVMObject *)iterator));
-    if (!iterator->body.hash_state.curr)
+
+    MVMStrHashTable *hashtable = &(((MVMHash *)iterator->body.target)->body.hashtable);
+
+#if HASH_DEBUG_ITER
+    struct MVMStrHashTableControl *control = hashtable->table;
+    MVMuint64 ht_id = control ? control->ht_id : 0;
+    if (iterator->body.hash_state.next.owner != ht_id) {
+        MVM_oops(tc, "MVM_itereky_s called with an iterator from a different hash table: %016" PRIx64 " != %016" PRIx64,
+                 iterator->body.hash_state.next.owner, ht_id);
+    }
+#endif
+
+    if (MVM_str_hash_at_end(tc, hashtable, iterator->body.hash_state.curr)
+        || MVM_str_hash_at_start(tc, hashtable, iterator->body.hash_state.curr))
         MVM_exception_throw_adhoc(tc, "You have not advanced to the first item of the hash iterator, or have gone past the end");
-    return MVM_HASH_KEY(iterator->body.hash_state.curr);
+
+    struct MVMHashEntry *entry = MVM_str_hash_current_nocheck(tc, hashtable, iterator->body.hash_state.curr);
+    return entry->hash_handle.key;
 }
 
 MVMObject * MVM_iterval(MVMThreadContext *tc, MVMIter *iterator) {
@@ -375,9 +311,22 @@ MVMObject * MVM_iterval(MVMThreadContext *tc, MVMIter *iterator) {
         REPR(target)->pos_funcs.at_pos(tc, STABLE(target), target, OBJECT_BODY(target), body->array_state.index, &result, MVM_reg_obj);
     }
     else if (iterator->body.mode == MVM_ITER_MODE_HASH) {
-        if (!iterator->body.hash_state.curr)
+        MVMStrHashTable *hashtable = &(((MVMHash *)iterator->body.target)->body.hashtable);
+
+#if HASH_DEBUG_ITER
+        struct MVMStrHashTableControl *control = hashtable->table;
+        MVMuint64 ht_id = control ? control->ht_id : 0;
+        if (iterator->body.hash_state.next.owner != ht_id) {
+        MVM_oops(tc, "MVM_iterval called with an iterator from a different hash table: %016" PRIx64 " != %016" PRIx64,
+                 iterator->body.hash_state.next.owner, ht_id);
+        }
+#endif
+
+        if (MVM_str_hash_at_end(tc, hashtable, iterator->body.hash_state.curr)
+            || MVM_str_hash_at_start(tc, hashtable, iterator->body.hash_state.curr))
             MVM_exception_throw_adhoc(tc, "You have not advanced to the first item of the hash iterator, or have gone past the end");
-        result.o = iterator->body.hash_state.curr->value;
+        struct MVMHashEntry *entry = MVM_str_hash_current_nocheck(tc, hashtable, iterator->body.hash_state.curr);
+        result.o = entry->value;
         if (!result.o)
             result.o = tc->instance->VMNull;
     }
