@@ -18,9 +18,7 @@ MVMCompUnit * MVM_cu_from_bytes(MVMThreadContext *tc, MVMuint8 *bytes, MVMuint32
     MVM_gc_allocate_gen2_default_clear(tc);
 
     /* Process the input. */
-    MVMROOT(tc, cu, {
-        MVM_bytecode_unpack(tc, cu);
-    });
+    MVM_bytecode_unpack(tc, cu);
 
     /* Resolve HLL config. It may contain nursery pointers, so fire write
      * barrier on it. */
@@ -31,48 +29,95 @@ MVMCompUnit * MVM_cu_from_bytes(MVMThreadContext *tc, MVMuint8 *bytes, MVMuint32
 }
 
 /* Loads a compilation unit from a bytecode file, mapping it into memory. */
-MVMCompUnit * MVM_cu_map_from_file(MVMThreadContext *tc, const char *filename) {
+MVMCompUnit * MVM_cu_map_from_file(MVMThreadContext *tc, const char *filename, MVMint32 free_filename) {
     MVMCompUnit *cu          = NULL;
     void        *block       = NULL;
     void        *handle      = NULL;
     uv_file      fd;
     MVMuint64    size;
     uv_fs_t req;
+    char *waste[2] = { free_filename ? (char *)filename : NULL, NULL };
+
+    unsigned long interval_id = MVM_telemetry_interval_start(tc, "map cu from file");
 
     /* Ensure the file exists, and get its size. */
-    if (uv_fs_stat(tc->loop, &req, filename, NULL) < 0) {
-        MVM_exception_throw_adhoc(tc, "While looking for '%s': %s", filename, uv_strerror(req.result));
+    if (uv_fs_stat(NULL, &req, filename, NULL) < 0) {
+        MVM_exception_throw_adhoc_free(tc, waste, "While looking for '%s': %s", filename, uv_strerror(req.result));
     }
 
     size = req.statbuf.st_size;
 
     /* Map the bytecode file into memory. */
-    if ((fd = uv_fs_open(tc->loop, &req, filename, O_RDONLY, 0, NULL)) < 0) {
-        MVM_exception_throw_adhoc(tc, "While trying to open '%s': %s", filename, uv_strerror(req.result));
+    if ((fd = uv_fs_open(NULL, &req, filename, O_RDONLY, 0, NULL)) < 0) {
+        MVM_exception_throw_adhoc_free(tc, waste, "While trying to open '%s': %s", filename, uv_strerror(req.result));
     }
 
+    MVM_telemetry_interval_annotate((uintptr_t)fd, interval_id, "this handle");
+    MVM_telemetry_interval_annotate((uintptr_t)size, interval_id, "this size");
+
     if ((block = MVM_platform_map_file(fd, &handle, (size_t)size, 0)) == NULL) {
         /* FIXME: check errno or GetLastError() */
-        MVM_exception_throw_adhoc(tc, "Could not map file '%s' into memory: %s", filename, "FIXME");
+        MVM_exception_throw_adhoc_free(tc, waste, "Could not map file '%s' into memory: %s", filename, "FIXME");
     }
 
-    if (uv_fs_close(tc->loop, &req, fd, NULL) < 0) {
-        MVM_exception_throw_adhoc(tc, "Failed to close filehandle: %s", uv_strerror(req.result));
+    if (uv_fs_close(NULL, &req, fd, NULL) < 0) {
+        MVM_exception_throw_adhoc_free(tc, waste, "Failed to close filehandle for '%s': %s", filename, uv_strerror(req.result));
     }
 
+    if (free_filename)
+        MVM_free((char *)filename);
+
     /* Turn it into a compilation unit. */
     cu = MVM_cu_from_bytes(tc, (MVMuint8 *)block, (MVMuint32)size);
     cu->body.handle = handle;
     cu->body.deallocate = MVM_DEALLOCATE_UNMAP;
+
+    MVM_telemetry_interval_stop(tc, interval_id, "done mapping");
+
+    return cu;
+}
+
+/* Loads a compilation unit from a bytecode file handle, mapping it into memory. */
+MVMCompUnit * MVM_cu_map_from_file_handle(MVMThreadContext *tc, uv_file fd, MVMuint64 pos) {
+    MVMCompUnit *cu          = NULL;
+    void        *block       = NULL;
+    void        *handle      = NULL;
+    MVMuint64    size;
+    uv_fs_t req;
+
+    unsigned long interval_id = MVM_telemetry_interval_start(tc, "map cu from file handle");
+    MVM_telemetry_interval_annotate((uintptr_t)handle, interval_id, "this handle");
+
+    /* Ensure the file exists, and get its size. */
+    if (uv_fs_fstat(NULL, &req, fd, NULL) < 0) {
+        MVM_exception_throw_adhoc(tc, "Trying to stat: %s", uv_strerror(req.result));
+    }
+
+    size = req.statbuf.st_size;
+
+    MVM_telemetry_interval_annotate((uintptr_t)size, interval_id, "this size");
+
+    if ((block = MVM_platform_map_file(fd, &handle, (size_t)size, 0)) == NULL) {
+        /* FIXME: check errno or GetLastError() */
+        MVM_exception_throw_adhoc(tc, "Could not map file into memory: %s", "FIXME");
+    }
+
+    block = ((char*)block) + pos;
+
+    /* Turn it into a compilation unit. */
+    cu = MVM_cu_from_bytes(tc, (MVMuint8 *)block, (MVMuint32)size);
+    cu->body.handle = handle;
+    cu->body.deallocate = MVM_DEALLOCATE_UNMAP;
+    MVM_telemetry_interval_stop(tc, interval_id, "done mapping");
     return cu;
 }
 
 /* Adds an extra callsite, needed due to an inlining, and returns its index. */
 MVMuint16 MVM_cu_callsite_add(MVMThreadContext *tc, MVMCompUnit *cu, MVMCallsite *cs) {
     MVMuint16 found = 0;
-    MVMuint16 idx;
+    MVMuint32 idx;
 
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    uv_mutex_lock(cu->body.inline_tweak_mutex);
 
     /* See if we already know this callsite. */
     for (idx = 0; idx < cu->body.num_callsites; idx++)
@@ -82,14 +127,19 @@ MVMuint16 MVM_cu_callsite_add(MVMThreadContext *tc, MVMCompUnit *cu, MVMCallsite
         }
     if (!found) {
         /* Not known; let's add it. */
+        size_t orig_size = cu->body.num_callsites * sizeof(MVMCallsite *);
+        size_t new_size = (cu->body.num_callsites + 1) * sizeof(MVMCallsite *);
+        MVMCallsite **new_callsites = MVM_malloc(new_size);
+        memcpy(new_callsites, cu->body.callsites, orig_size);
         idx = cu->body.num_callsites;
-        cu->body.callsites = MVM_realloc(cu->body.callsites,
-            (idx + 1) * sizeof(MVMCallsite *));
-        cu->body.callsites[idx] = MVM_callsite_copy(tc, cs);
+        new_callsites[idx] = cs->is_interned ? cs : MVM_callsite_copy(tc, cs);
+        if (cu->body.callsites)
+            MVM_free_at_safepoint(tc, cu->body.callsites);
+        cu->body.callsites = new_callsites;
         cu->body.num_callsites++;
     }
 
-    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    uv_mutex_unlock(cu->body.inline_tweak_mutex);
 
     return idx;
 }
@@ -99,7 +149,7 @@ MVMuint32 MVM_cu_string_add(MVMThreadContext *tc, MVMCompUnit *cu, MVMString *st
     MVMuint32 found = 0;
     MVMuint32 idx;
 
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    uv_mutex_lock(cu->body.inline_tweak_mutex);
 
     /* See if we already know this string; only consider those added already by
      * inline, since we don't intern and don't want this to be costly to hunt. */
@@ -110,14 +160,19 @@ MVMuint32 MVM_cu_string_add(MVMThreadContext *tc, MVMCompUnit *cu, MVMString *st
         }
     if (!found) {
         /* Not known; let's add it. */
+        size_t orig_size = cu->body.num_strings * sizeof(MVMString *);
+        size_t new_size = (cu->body.num_strings + 1) * sizeof(MVMString *);
+        MVMString **new_strings = MVM_malloc(new_size);
+        memcpy(new_strings, cu->body.strings, orig_size);
         idx = cu->body.num_strings;
-        cu->body.strings = MVM_realloc(cu->body.strings,
-            (idx + 1) * sizeof(MVMString *));
-        cu->body.strings[idx] = str;
+        new_strings[idx] = str;
+        if (cu->body.strings)
+            MVM_free_at_safepoint(tc, cu->body.strings);
+        cu->body.strings = new_strings;
         cu->body.num_strings++;
     }
 
-    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    uv_mutex_unlock(cu->body.inline_tweak_mutex);
 
     return idx;
 }
