@@ -1,15 +1,16 @@
 #include "moar.h"
 
 int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
-    MVMDLLRegistry *entry;
     char *cpath;
     DLLib *lib;
 
-    MVM_string_flatten(tc, name);
+    if (!MVM_str_hash_key_is_valid(tc, name)) {
+        MVM_str_hash_key_throw_invalid(tc, name);
+    }
 
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
-    MVM_HASH_GET(tc, tc->instance->dll_registry, name, entry);
+    MVMDLLRegistry *entry = MVM_fixkey_hash_fetch_nocheck(tc, &tc->instance->dll_registry, name);
 
     /* already loaded */
     if (entry && entry->lib) {
@@ -17,13 +18,11 @@ int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
         return 0;
     }
 
-    MVMROOT(tc, name, {
-        MVMROOT(tc, path, {
-            path = MVM_file_in_libpath(tc, path);
-        });
-    });
+    MVMROOT2(tc, name, path) {
+        path = MVM_file_in_libpath(tc, path);
+    }
 
-    cpath = MVM_string_utf8_encode_C_string(tc, path);
+    cpath = MVM_string_utf8_c8_encode_C_string(tc, path);
     lib = MVM_nativecall_load_lib(cpath);
 
     if (!lib) {
@@ -35,12 +34,10 @@ int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
     MVM_free(cpath);
 
     if (!entry) {
-        entry = MVM_malloc(sizeof *entry);
-        entry->name = name;
+        entry = MVM_fixkey_hash_insert_nocheck(tc, &tc->instance->dll_registry, name);
         entry->refcount = 0;
-
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&entry->name);
-        MVM_HASH_BIND(tc, tc->instance->dll_registry, name, entry);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&entry->hash_key,
+            "DLL name hash key");
     }
 
     entry->lib = lib;
@@ -51,16 +48,19 @@ int MVM_dll_load(MVMThreadContext *tc, MVMString *name, MVMString *path) {
 }
 
 int MVM_dll_free(MVMThreadContext *tc, MVMString *name) {
-    MVMDLLRegistry *entry;
+    if (!MVM_str_hash_key_is_valid(tc, name)) {
+        MVM_str_hash_key_throw_invalid(tc, name);
+    }
 
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
-    MVM_string_flatten(tc, name);
-    MVM_HASH_GET(tc, tc->instance->dll_registry, name, entry);
+    MVMDLLRegistry *entry = MVM_fixkey_hash_fetch_nocheck(tc, &tc->instance->dll_registry, name);
 
     if (!entry) {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
         uv_mutex_unlock(&tc->instance->mutex_dll_registry);
-        MVM_exception_throw_adhoc(tc, "cannot free non-existent library");
+        MVM_exception_throw_adhoc_free(tc, waste, "cannot free non-existent library '%s'", c_name);
     }
 
     /* already freed */
@@ -70,8 +70,10 @@ int MVM_dll_free(MVMThreadContext *tc, MVMString *name) {
     }
 
     if (entry->refcount > 0) {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
         uv_mutex_unlock(&tc->instance->mutex_dll_registry);
-        MVM_exception_throw_adhoc(tc, "cannot free in-use library");
+        MVM_exception_throw_adhoc_free(tc, waste, "cannot free in-use library '%s'", c_name);
     }
 
     MVM_nativecall_free_lib(entry->lib);
@@ -84,29 +86,35 @@ int MVM_dll_free(MVMThreadContext *tc, MVMString *name) {
 
 MVMObject * MVM_dll_find_symbol(MVMThreadContext *tc, MVMString *lib,
         MVMString *sym) {
-    MVMDLLRegistry *entry;
     MVMDLLSym *obj;
     char *csym;
     void *address;
 
+    if (!MVM_str_hash_key_is_valid(tc, lib)) {
+        MVM_str_hash_key_throw_invalid(tc, lib);
+    }
+
     uv_mutex_lock(&tc->instance->mutex_dll_registry);
 
-    MVM_string_flatten(tc, lib);
-    MVM_HASH_GET(tc, tc->instance->dll_registry, lib, entry);
+    MVMDLLRegistry *entry = MVM_fixkey_hash_fetch_nocheck(tc, &tc->instance->dll_registry, lib);
 
     if (!entry) {
+        char *c_lib = MVM_string_utf8_encode_C_string(tc, lib);
+        char *waste[] = { c_lib, NULL };
         uv_mutex_unlock(&tc->instance->mutex_dll_registry);
-        MVM_exception_throw_adhoc(tc,
-                "cannot find symbol in non-existent library");
+        MVM_exception_throw_adhoc_free(tc, waste,
+                "cannot find symbol '%s' in non-existent library", c_lib);
     }
 
     if (!entry->lib) {
+        char *c_lib = MVM_string_utf8_encode_C_string(tc, lib);
+        char *waste[] = { c_lib, NULL };
         uv_mutex_unlock(&tc->instance->mutex_dll_registry);
-        MVM_exception_throw_adhoc(tc,
-                "cannot find symbol in unloaded library");
+        MVM_exception_throw_adhoc_free(tc, waste,
+                "cannot find symbol '%s' in unloaded library", c_lib);
     }
 
-    csym = MVM_string_utf8_encode_C_string(tc, sym);
+    csym = MVM_string_utf8_c8_encode_C_string(tc, sym);
     address = MVM_nativecall_find_sym(entry->lib, csym);
     MVM_free(csym);
 
